// Adapted from: Jos Stam, "Real-Time Fluid Dynamics for Games". Proceedings of the Game Developer Conference, March 2003.
// For details on how the individual functions work, please consult the above publication.

#include "StamFluidComponent.h"
#include "Engine/Canvas.h"
#include "Engine/VolumeTexture.h"

DEFINE_LOG_CATEGORY(LogStam);

UStamFluidComponent::UStamFluidComponent()
{
	PrimaryComponentTick.bCanEverTick = true;
}

void UStamFluidComponent::InjectFluid(FVector position, float amount)
{
	// Map the unit vector to a specific cell.
	int x = position.X * Size;
	int y = position.Y * Size;
	int z = position.Z * Size;

	x = std::min(Size, std::max(1, x));
	y = std::min(Size, std::max(1, y));
	z = std::min(Size, std::max(1, z));

	Density_prev[GetIndex(x, y, z)] += amount;
}

void UStamFluidComponent::ApplyForce(FVector position, FVector force)
{
	int x = position.X * Size;
	int y = position.Y * Size;
	int z = position.Z * Size;

	x = std::min(Size, std::max(1, x));
	y = std::min(Size, std::max(1, y));
	z = std::min(Size, std::max(1, z));

	VelocityX_prev[GetIndex(x, y, z)] = force.X;
	VelocityY_prev[GetIndex(x, y, z)] = force.Y;
	VelocityZ_prev[GetIndex(x, y, z)] = force.Z;
}

// Called when the game starts
void UStamFluidComponent::BeginPlay()
{
	Super::BeginPlay();

	CreateVolumeTransient();

	// Expand the actual computation grid to account for edge cases
	DensityGridSize = Size + 2;

	// Create data arrays, fill them with zeroes
	Density.Init(0, DensityGridSize * DensityGridSize * DensityGridSize);
	Density_prev.Init(0, DensityGridSize * DensityGridSize * DensityGridSize);

	VelocityX.Init(0, DensityGridSize * DensityGridSize * DensityGridSize);
	VelocityX_prev.Init(0, DensityGridSize * DensityGridSize * DensityGridSize);
	
	VelocityY.Init(0, DensityGridSize * DensityGridSize * DensityGridSize);
	VelocityY_prev.Init(0, DensityGridSize * DensityGridSize * DensityGridSize);
	
	VelocityZ.Init(0, DensityGridSize * DensityGridSize * DensityGridSize);
	VelocityZ_prev.Init(0, DensityGridSize * DensityGridSize * DensityGridSize);
}


// Called every frame
void UStamFluidComponent::TickComponent(float DeltaTime, ELevelTick TickType, FActorComponentTickFunction* ThisTickFunction)
{
	Super::TickComponent(DeltaTime, TickType, ThisTickFunction);

	// Run Stam's stable fluid solver
	if (Simulating)
	{
		VelocityStep(DeltaTime);
		DensityStep(DeltaTime);
	}

	UE_LOG(LogStam, Warning, TEXT("Value at center: %f"), Density[GetIndex(Size / 2 + 1, Size / 2 + 1, Size / 2 + 1)]);

	// Lock texture data so we can update it
	uint8* ByteArray = (uint8*)VolumeTexture->PlatformData->Mips[0].BulkData.Lock(LOCK_READ_WRITE);

	// Loop through all our cells and convert their values to color.
	// Only takes into account values between 0.0-1.0 and maps them to 0-255
	int cur = 0;
	for (int i = 1; i <= Size; i++)
	{
		for (int j = 1; j <= Size; j++)
		{
			for (int k = 1; k <= Size; k++)
			{
				int val = 255 * Density[GetIndex(k, j, i)] / 1.0f;
				
				val = std::min(255, std::max(0, val));

				// 4 values: R G B A, a byte each
				ByteArray[cur + 0] = val;
				ByteArray[cur + 1] = val;
				ByteArray[cur + 2] = val;
				ByteArray[cur + 3] = val;

				cur += 4;
			}
		}
	}

	// We're done updating - unlock the data and tell UE to update the texture on the GPU side.
	VolumeTexture->PlatformData->Mips[0].BulkData.Unlock();
	VolumeTexture->UpdateResource();

	// Clear data for next tick.
	FMemory::Memset(Density_prev.GetData(), 0, Density_prev.GetAllocatedSize());
	FMemory::Memset(VelocityX_prev.GetData(), 0, VelocityX_prev.GetAllocatedSize());
	FMemory::Memset(VelocityY_prev.GetData(), 0, VelocityY_prev.GetAllocatedSize());
	FMemory::Memset(VelocityZ_prev.GetData(), 0, VelocityZ_prev.GetAllocatedSize());
}

void UStamFluidComponent::CreateVolumeTransient()
{
	// First create the object itself.
	VolumeTexture = NewObject<UVolumeTexture>(GetTransientPackage(), NAME_None, RF_Transient);
	
	if (!VolumeTexture->PlatformData)
	{
		VolumeTexture->PlatformData = new FTexturePlatformData();
	}
	// Set Dimensions and Pixel format.
	VolumeTexture->PlatformData->SizeX = Size;
	VolumeTexture->PlatformData->SizeY = Size;
	VolumeTexture->PlatformData->SetNumSlices(Size);
	VolumeTexture->PlatformData->PixelFormat = EPixelFormat::PF_R8G8B8A8;
	// Set sRGB and streaming to false.
	VolumeTexture->SRGB = false;
	VolumeTexture->NeverStream = true;
	
	int PixelByteSize = 8 * 4;
	const long long TotalSize = (long long) Size * Size * Size * PixelByteSize;

	// Create the one and only mip in this texture.
	FTexture2DMipMap* mip = new FTexture2DMipMap();
	mip->SizeX = Size;
	mip->SizeY = Size;
	mip->SizeZ = Size;

	mip->BulkData.Lock(LOCK_READ_WRITE);
	// Allocate memory in the mip and copy the actual texture data inside
	uint8* ByteArray = (uint8*) mip->BulkData.Realloc(TotalSize);

	// Memset to zero
	FMemory::Memset(ByteArray, 0, TotalSize);

	mip->BulkData.Unlock();

	// Newly created Volume textures have this null'd
	if (!VolumeTexture->PlatformData)
	{
		VolumeTexture->PlatformData = new FTexturePlatformData();
	}
	// Add the new MIP to the list of mips.
	VolumeTexture->PlatformData->Mips.Add(mip);

	// Tell Unreal to update the rest.
	VolumeTexture->UpdateResource();
}

void UStamFluidComponent::AddSources(TArray<float>& x, const TArray<float>& s, float DeltaTime)
{
	int size = DensityGridSize * DensityGridSize * DensityGridSize;
	for (int i = 0; i < size; i++)
	{
		x[i] += s[i] * DeltaTime; 
	}
}

void UStamFluidComponent::Diffuse(int btype, TArray<float>& x, 
	TArray<float>& x0, float diff, float DeltaTime)
{
	// The actual factor is the input one scaled by the time step and scaled to the size of our simulation grid.
	float a = DeltaTime * diff * Size * Size * Size;
 
	for (int st = 0; st < LinearSolverSteps; st++)
	{ 
		for (int i = 1; i <= Size; i++)
		{ 
			for (int j = 1; j <= Size; j++)
			{
				for (int k = 1; k <= Size; k++)
				{
				    x[GetIndex(i,j,k)] = (x0[GetIndex(i,j,k)] + a*(
				    	x[GetIndex(i-1,j,k)]+x[GetIndex(i+1,j,k)]+ 
				    	x[GetIndex(i,j-1,k)]+x[GetIndex(i,j+1,k)]+
				    	x[GetIndex(i,j,k-1)]+x[GetIndex(i,j,k+1)]))/(1+6*a);
				}
			} 
		}
		SetBounds(btype, x); 
	}
}

void UStamFluidComponent::DiffuseBad(int btype, TArray<float>& x, 
	TArray<float>& x0, float diff, float DeltaTime)
{
	float a = DeltaTime * diff * Size * Size * Size;
 
	for (int i = 1; i <= Size; i++)
	{ 
		for (int j = 1; j <= Size; j++)
		{
			for (int k = 1; k <= Size; k++)
			{
				// Naive forward approach. Faster than the stable solution and works for sufficiently small
				// time steps (and diffusion factors), but can easily become unstable (imagine a pause elsewhere in the
				// game, resulting in a very high time step).
				x[GetIndex(i, j, k)] = x0[GetIndex(i, j, k)]
				+ a*(x0[GetIndex(i + 1, j, k)]
				+ x0[GetIndex(i - 1, j, k)]
				+ x0[GetIndex(i, j + 1, k)]
				+ x0[GetIndex(i, j - 1, k)]
				+ x0[GetIndex(i, j, k + 1)]
				+ x0[GetIndex(i, j, k - 1)]
				- 6 * x0[GetIndex(i, j, k)]);
			}
		}
	}
	SetBounds(btype, x);
}

void UStamFluidComponent::Advect(int btype, TArray<float>& d, TArray<float>& d0, TArray<float>& vx, TArray<float>& vy, TArray<float>& vz, float DeltaTime)
{
	float dt0 = DeltaTime * Size;
	for (int i = 1; i <= Size; i++)
	{
		for (int j = 1; j <= Size; j++)
		{
			for (int k = 1; k <= Size; k++)
			{
				// Find source position
				float x = i - dt0 * vx[GetIndex(i, j, k)];
				float y = j - dt0 * vy[GetIndex(i, j, k)];
				float z = k - dt0 * vz[GetIndex(i, j, k)];

				// Make sure we're sufficiently inside the grid
				if (x < 0.5f) x = 0.5f; if (x > Size + 0.5f) x = Size + 0.5f;
				if (y < 0.5f) y = 0.5f; if (y > Size + 0.5f) y = Size + 0.5f;
				if (z < 0.5f) z = 0.5f; if (z > Size + 0.5f) z = Size + 0.5f;

				// Compute neighbouring cells and their contributions
				int i0 = (int)x;
				int i1 = i0 + 1;
				
				int j0 = (int)y;
				int j1 = j0 + 1;
				
				int k0 = (int)z;
				int k1 = k0 + 1;

				float s1 = x - i0;
				float s0 = 1 - s1;
				
				float t1 = y - j0;
				float t0 = 1 - t1;
				
				float u1 = z - k0;
				float u0 = 1 - u1;

				// Linearly interpolate between the 8 neighbours
				d[GetIndex(i, j, k)] =
					s0 * (t0 * (u0 * d0[GetIndex(i0, j0, k0)] + u1 * d0[GetIndex(i0, j0, k1)]) +
						  t1 * (u0 * d0[GetIndex(i0, j1, k0)] + u1 * d0[GetIndex(i0, j1, k1)])) +
					s1 * (t0 * (u0 * d0[GetIndex(i1, j0, k0)] + u1 * d0[GetIndex(i1, j0, k1)]) +
						  t1 * (u0 * d0[GetIndex(i1, j1, k0)] + u1 * d0[GetIndex(i1, j1, k1)]));
			}
		}
	}
	SetBounds(btype, d);
}

void UStamFluidComponent::DensityStep(float DeltaTime)
{
	// Stam's original code uses pointer swapping - here we just swap the variable names instead.
	// Swapping is used instead of additional temporary arrays. Swapping means result is always computed to a "new" array.
	AddSources(Density, Density_prev, DeltaTime);
	// Try both variants with different parameters, see how the bad variant breaks quickly with higher values.
	Diffuse(0, Density_prev, Density, DiffusionFactor, DeltaTime);
	//DiffuseBad(0, Density_prev, Density, DiffusionFactor, DeltaTime);
	Advect(0, Density, Density_prev, VelocityX, VelocityY, VelocityZ, DeltaTime);
}

void UStamFluidComponent::Project(TArray<float>& vx, TArray<float>& vy, TArray<float>& vz, TArray<float>& p,
	TArray<float>& div)
{
	float h = 1.0f/Size; 
	for (int i = 1; i <= Size; i++)
	{ 
		for (int j = 1; j <= Size; j++)
		{
			for (int k = 1; k <= Size; k++)
			{
				div[GetIndex(i, j, k)] = -1.0f / 3.0f * h * (
					vx[GetIndex(i + 1, j, k)] - vx[GetIndex(i - 1, j, k)] +
					vy[GetIndex(i, j + 1, k)] - vy[GetIndex(i, j - 1, k)] +
					vz[GetIndex(i, j, k + 1) - vz[GetIndex(i, j, k - 1)]]);
				p[GetIndex(i, j, k)] = 0;
			}
		} 
	}
	SetBounds(0, div); SetBounds(0, p);

	// The gradient field is found using the Gauss-Seidel method again.
	for (int step = 0; step < LinearSolverSteps; step++)
	{ 
		for (int i = 1; i <= Size; i++)
		{
			for (int j = 1; j <= Size; j++)
			{
				for (int k = 1; k <= Size; k++)
				{
					p[GetIndex(i, j, k)] = (div[GetIndex(i, j, k)] +
						p[GetIndex(i - 1, j, k)] +
						p[GetIndex(i + 1, j, k)] +
						p[GetIndex(i, j + 1, k)] +
						p[GetIndex(i, j - 1, k)] +
						p[GetIndex(i, j, k + 1)] +
						p[GetIndex(i, j, k - 1)]) / 6;	
				} 
			}
		}
		SetBounds(0, p);
	} 
 
	for (int i = 1; i <= Size; i++)
	{
		for (int j = 1;j <= Size; j++)
		{
			for (int k = 1; k <= Size; k++)
			{
				vx[GetIndex(i, j, k)] -= 0.5f*(p[GetIndex(i + 1, j, k)] - p[GetIndex(i - 1, j, k)])/h;
				vy[GetIndex(i, j, k)] -= 0.5f*(p[GetIndex(i, j + 1, k)] - p[GetIndex(i, j - 1, k)])/h;
				vz[GetIndex(i, j, k)] -= 0.5f*(p[GetIndex(i, j, k + 1)] - p[GetIndex(i, j, k - 1)])/h;
			}
		} 
	}
	SetBounds(1, vx); SetBounds(2, vy); SetBounds(3, vz);
}

void UStamFluidComponent::VelocityStep(float DeltaTime)
{
	// Stam's original code uses pointer swapping - here we just swap the variable names instead.
	AddSources(VelocityX, VelocityX_prev, DeltaTime);
	AddSources(VelocityY, VelocityY_prev, DeltaTime);
	AddSources(VelocityZ, VelocityZ_prev, DeltaTime);

	Diffuse(1, VelocityX_prev, VelocityX, Viscosity, DeltaTime);
	Diffuse(2, VelocityY_prev, VelocityY, Viscosity, DeltaTime);
	Diffuse(3, VelocityZ_prev, VelocityZ, Viscosity, DeltaTime);

	Project(VelocityX_prev, VelocityY_prev, VelocityZ_prev, VelocityX, VelocityY);
	
	Advect(1, VelocityX, VelocityX_prev, VelocityX_prev, VelocityY_prev, VelocityZ_prev, DeltaTime);
	Advect(2, VelocityY, VelocityY_prev, VelocityX_prev, VelocityY_prev, VelocityZ_prev, DeltaTime);
	Advect(3, VelocityZ, VelocityZ_prev, VelocityX_prev, VelocityY_prev, VelocityZ_prev, DeltaTime);

	Project(VelocityX, VelocityY, VelocityZ, VelocityX_prev, VelocityY_prev);
}

void UStamFluidComponent::SetBounds(int type, TArray<float>& x)
{
	const int l = Size + 1;
	
	for (int i = 1; i <= Size; i++)
	{
		// First take care of the walls of the fluid's bounding box
		// The type parameter value is used here to determine when to "reflect" the vector along the wall
		for (int j = 1; j <= Size; j++)
		{
			x[GetIndex(0, i, j)] = type == 1 ? -x[GetIndex(1, i, j)] : x[GetIndex(1, i, j)];
			x[GetIndex(l, i, j)] = type == 1 ? -x[GetIndex(l - 1, i, j)] : x[GetIndex(l - 1, i, j)];
			
			x[GetIndex(i, 0, j)] = type == 2 ? -x[GetIndex(i, 1, j)] : x[GetIndex(i, 1, j)];
			x[GetIndex(i, l, j)] = type == 2 ? -x[GetIndex(i, l - 1, j)] : x[GetIndex(i, l - 1, j)];
			
			x[GetIndex(i, j, 0)] = type == 3 ? -x[GetIndex(i, j, 1)] : x[GetIndex(i, j, 1)];
			x[GetIndex(i, j, l)] = type == 3 ? -x[GetIndex(i, j, l - 1)] : x[GetIndex(i, j, l - 1)];
		}

		// Next set limiting values along the 6 edges of the box
		x[GetIndex(0, i, l)] = 0.5f * (x[GetIndex(1    , i, l)] + x[GetIndex(0, i, l - 1)]);
		x[GetIndex(l, i, l)] = 0.5f * (x[GetIndex(l - 1, i, l)] + x[GetIndex(l, i, l - 1)]);
		x[GetIndex(0, i, 0)] = 0.5f * (x[GetIndex(1    , i, 0)] + x[GetIndex(0, i, 1    )]);
		x[GetIndex(l, i, 0)] = 0.5f * (x[GetIndex(l - 1, i, 0)] + x[GetIndex(l, i, 1    )]);
		
		x[GetIndex(i, 0, l)] = 0.5f * (x[GetIndex(i, 1    , l)] + x[GetIndex(i, 0, l - 1)]);
		x[GetIndex(i, l, l)] = 0.5f * (x[GetIndex(i, l - 1, l)] + x[GetIndex(i, l, l - 1)]);
		x[GetIndex(i, 0, 0)] = 0.5f * (x[GetIndex(i, 1    , 0)] + x[GetIndex(i, 0, 1    )]);
		x[GetIndex(i, l, 0)] = 0.5f * (x[GetIndex(i, l - 1, 0)] + x[GetIndex(i, l, 1    )]);
		
		x[GetIndex(0, l, i)] = 0.5f * (x[GetIndex(1    , l, i)] + x[GetIndex(0, l - 1, i)]);
		x[GetIndex(l, l, i)] = 0.5f * (x[GetIndex(l - 1, l, i)] + x[GetIndex(l, l - 1, i)]);
		x[GetIndex(0, 0, i)] = 0.5f * (x[GetIndex(1    , 0, i)] + x[GetIndex(0, 0    , i)]);
		x[GetIndex(l, 0, i)] = 0.5f * (x[GetIndex(l - 1, 0, i)] + x[GetIndex(l, 0    , i)]);
	}

	// Finally handle the 8 corners of the box
	constexpr float third = 1.0 / 3.0;
	x[GetIndex(0, 0, 0)] = third * (x[GetIndex(1    , 0, 0)] + x[GetIndex(0, 1    , 0)] + x[GetIndex(0, 0, 1    )]);
	x[GetIndex(0, l, 0)] = third * (x[GetIndex(1    , l, 0)] + x[GetIndex(0, l - 1, 0)] + x[GetIndex(0, l, 1    )]);
	x[GetIndex(l, 0, 0)] = third * (x[GetIndex(l - 1, 0, 0)] + x[GetIndex(l, 1    , 0)] + x[GetIndex(l, 0, 1    )]);
	x[GetIndex(l, l, 0)] = third * (x[GetIndex(l - 1, l, 0)] + x[GetIndex(l, l - 1, 0)] + x[GetIndex(l, l, 1    )]);
	x[GetIndex(0, 0, l)] = third * (x[GetIndex(1    , 0, l)] + x[GetIndex(0, 1    , l)] + x[GetIndex(0, 0, l - 1)]);
	x[GetIndex(0, l, l)] = third * (x[GetIndex(1    , l, l)] + x[GetIndex(0, l - 1, l)] + x[GetIndex(0, l, l - 1)]);
	x[GetIndex(l, 0, l)] = third * (x[GetIndex(l - 1, 0, l)] + x[GetIndex(l, 1    , l)] + x[GetIndex(l, 0, l - 1)]);
	x[GetIndex(l, l, l)] = third * (x[GetIndex(l - 1, l, l)] + x[GetIndex(l, l - 1, l)] + x[GetIndex(l, l, l - 1)]);
}

