#pragma once

#include "CoreMinimal.h"
#include "Components/ActorComponent.h"
#include "Engine/TextureRenderTargetVolume.h"
#include "StamFluidComponent.generated.h"

DECLARE_LOG_CATEGORY_EXTERN(LogStam, Warning, All);

UCLASS( ClassGroup=(Custom), meta=(BlueprintSpawnableComponent) )
class CVICENI07_API UStamFluidComponent : public UActorComponent
{
	GENERATED_BODY()

public:	
	UStamFluidComponent();

	UPROPERTY(EditDefaultsOnly, Category="Volume Parameters")
	int Size = 32;

	UPROPERTY(BlueprintReadOnly, Category="Volume Parameters")
	UVolumeTexture* VolumeTexture;

	UPROPERTY(EditAnywhere, Category="Fluid Simulation")
	int LinearSolverSteps = 20;
	UPROPERTY(EditAnywhere, Category="Fluid Simulation")
	float DiffusionFactor = 0.0f;
	UPROPERTY(EditAnywhere, Category="Fluid Simulation")
	float Viscosity = 0.0f;
	UPROPERTY(EditAnywhere, Category="Fluid Simulation")
	bool Simulating = true;

	UFUNCTION(BlueprintCallable, Category="Fluid Simulation")
	// Injects new fluid into the simulation at the given position (expects a unit vector)
	void InjectFluid(FVector position, float amount);

	UFUNCTION(BlueprintCallable, Category="Fluid Simulation")
	// Adds a force into the simulation's force field (expects a unit vector)
	void ApplyForce(FVector position, FVector force);
	
protected:
	virtual void BeginPlay() override;

public:	
	virtual void TickComponent(float DeltaTime, ELevelTick TickType, FActorComponentTickFunction* ThisTickFunction) override;

private:
	/**
	 * @brief Creates a runtime VolumeTexture through Unreal API. This texture exists only when playing the game. It
	 * has a single mip level with the full resolution of our simulation cell grid (minus the edge cells).
	 */
	void CreateVolumeTransient();

	/**
	 * @brief Converts a set of coordinates to a cell array index. Simplifies working with a single-dimensional array
	 * representing a multi-dimensional structure.
	 * @param x X coordinate
	 * @param y Y coordinate
	 * @param z Z coordinate
	 * @return Resulting array index
	 */
	uint32 GetIndex(int x, int y, int z) const { return DensityGridSize * DensityGridSize * z + DensityGridSize * y + x; }

	
	/**
	 * @brief Add time-scaled values from the source array s to the destination array x.
	 * @param x Target array to add the sources to
	 * @param s Source array containing pre-second values to add
	 * @param DeltaTime Simulation time step
	 */
	void AddSources(TArray<float>& x, const TArray<float>& s, float DeltaTime);
	
	/**
	 * @brief Diffuses the values in the source array. Density/velocity each cell has is shared with its neighbours to
	 * simulate natural diffusion of gasses (and the force fields which affect them, i.e. wind). Used several times in
	 * the simulation. This version computes what density each cell should have if diffusion happened in reverse. This
	 * way each cell is a separate variable, resulting in a set of linear equations which can be solved using the
	 * iterative Gauss-Seidel method.
	 * @param btype Type of border limiting values to apply (used for velocities) - passed when calling SetBounds
	 * @param x Destination array - will contain the resulting diffused values
	 * @param x0 Source array containing the original densities/velocities to diffuse
	 * @param diff Diffusion factor - higher value means more diffusion
	 * @param DeltaTime Simulation time step
	 */
	void Diffuse(int btype, TArray<float>& x, TArray<float>& x0, float diff, float DeltaTime);

	
	/**
	 * @brief Same as Diffuse, but calculates diffusion forward in time. The algorithm is more straightforward but is
	 * very unstable, especially for larger diffusion factor and time step.
	 * @param btype Type of border limiting values to apply (used for velocities) - passed when calling SetBounds
	 * @param x Destination array - will contain the resulting diffused values
	 * @param x0 Source array containing the original densities/velocities to diffuse
	 * @param diff Diffusion factor - higher value means more diffusion
	 * @param DeltaTime Simulation time step
	 */
	void DiffuseBad(int btype, TArray<float>& x, TArray<float>& x0, float diff, float DeltaTime);

	
	/**
	 * @brief Computes the flow of the densities/forces due to advection from the vector field. This makes the fluid
	 * move across the cells. Computation is done similarly to the Diffusion step - backwards in time. We look at a position
	 * and linearly work out which cell the value came from. Since the resulting value is linearly interpolated from the
	 * computed position's neighbouring cells, this step also does pseudo-diffusion.
	 * @param btype Type of border limiting values to apply (used for velocities) - passed when calling SetBounds
	 * @param d Destination array - will contain the resulting values after the advection step
	 * @param d0 Source array containing densities/velocity component which will be moved
	 * @param vx Array of velocities - X axis vector component
	 * @param vy Array of velocities - Y axis vector component
	 * @param vz Array of velocities - Z axis vector component
	 * @param DeltaTime Simulation time step
	 */
	void Advect(int btype, TArray<float>& d, TArray<float>& d0, TArray<float>& vx, TArray<float>& vy, TArray<float>& vz, float DeltaTime);

	
	/**
	 * @brief High-level part of the simulation which uses the above functions to compute density changes in a single
	 * step of the simulation.
	 * @param DeltaTime Simulation time step
	 */
	void DensityStep(float DeltaTime);

	
	/**
	 * @brief This step ensures conservation of mass in our fluid's velocity field. This is done by using Hodge decomposition.
	 * First, a gradient is computed for the velocity field (into p), which is subtracted from our velocity field.
	 * The result is a mass conserving field which is saved to the input velocity field arrays.
	 * @param vx Array of velocities - X axis component
	 * @param vy Array of velocities - Y axis component
	 * @param vz Array of velocities - Z axis component
	 * @param p Temporary array used for computation. Needs to be the same size as the input arrays.
	 * @param div Temporary array used for computation. Needs to be the same size as the input arrays.
	 */
	void Project(TArray<float>& vx, TArray<float>& vy, TArray<float>& vz, TArray<float>& p, TArray<float>& div);

	
	/**
	 * @brief High-level part of the simulation which uses the above functions to compute velocity field changes in a
	 * single step of the simulation.
	 * @param DeltaTime Simulation time step
	 */
	void VelocityStep(float DeltaTime);

	/**
	 * @brief Applies edge conditions to the additional grid layer which is used in this simulation. This implementation
	 * encloses the fluid in a solid box.
	 * @param type Type of edge binding conditions to apply
	 * @param x Array of values to apply the conditions on
	 */
	void SetBounds(int type, TArray<float>& x);

	uint32 DensityGridSize;
	TArray<float> Density;
	TArray<float> Density_prev;

	TArray<float> VelocityX;
	TArray<float> VelocityY;
	TArray<float> VelocityZ;
	
	TArray<float> VelocityX_prev;
	TArray<float> VelocityY_prev;
	TArray<float> VelocityZ_prev;
};
